/* Diana.java
 *
 * created: Fri Oct  9 1998
 *
 * This file is part of Artemis
 * 
 * Copyright (C) 1998,1999  Genome Research Limited
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * $Header: /nfs/disk222/yeastpub/Repository/powmap/Diana.java,v 1.60 1999/04/21 15:37:40 kmr Exp $
 */

import diana.components.*;
import diana.sequence.*;
import diana.*;

import nsdb.EmblSeq;
import type.NoResult;
import nsdb.EmblPackage.Superceded;

import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.net.*;

/**
 *  This class contains the method main () for running Diana as a standalone
 *  program.  This class is also used by DianaApplet as a storage place for
 *  the global Clipboard object and the global Options object.
 *
 *  @author Kim Rutherford
 *  @version $Id: Diana.java,v 1.60 1999/04/21 15:37:40 kmr Exp $
 * */

public class Diana extends Frame {
  /**
   *  Main entry point for the stand-alone Diana.
   **/
  public static void main (final String [] args) {
    final String install_path =
      System.getProperties ().getProperty ("diana.install.path");

    if (install_path == null || install_path.length () == 0) {
      throw new Error ("the diana.install.path property was not set on the " +
                       "command line");
    } else {
      setCodeDirectory (new FileDocument (new File (install_path)));
    }

    try {
      final StringVector extra_keys =
        getOptions ().getOptionValues ("extra_keys");

      final embl.QualifierInfoVector extra_qualifiers =
        getOptions ().getExtraQualifiers ();

      embl.EntryInformation.initialise (getCodeDirectory (),
                                        extra_keys,
                                        extra_qualifiers);
    } catch (IOException e) {
      System.out.println ("could not initialise the embl package: " +
                          e.getMessage ());
      System.exit (1);
    } catch (embl.QualifierInfoException e) {
      System.out.println ("could not initialise the embl package: " +
                          e.getMessage ());
      System.exit (1);
    }

    main_window = new Diana ();

    main_window.setVisible (true);

    // read the entries given on the command line and in the diana.ini file
    main_window.readArgsAndOptions (args);
  }

  /**
   *  Read the entries given on the command line and in the diana.ini file.
   **/
  private void readArgsAndOptions (final String [] args) {
    // open the entries given in the options file (diana.ini)

    if (args.length == 0) {
      readDefaultEntries ();
    }

    EntryEdit last_entry_edit = null;

    for (int i = 0 ; i < args.length ; ++i) {
      final String new_entry_name = args[i];

//        System.out.println ("token " + new_entry_name);

      if (new_entry_name.startsWith ("+") && last_entry_edit != null) {
//          System.out.println ("adding " + new_entry_name);

        final embl.Entry new_embl_entry =
          openEntry (new_entry_name.substring (1));

        if (new_entry_name == null) {
          // the read failed
          break;
        }

        final Entry new_entry =
          new Entry (last_entry_edit.getEntryGroup ().getBases (),
                     new_embl_entry);

        last_entry_edit.getEntryGroup ().add (new_entry);

      } else {
        final embl.Entry new_embl_entry = openEntry (new_entry_name);

        if (new_embl_entry == null) {
          // the read failed
          break;
        }

        final Bases embl_bases = new Bases (new_embl_entry.getSequence ());

        final EntryGroup entry_group = new EntryGroup (embl_bases);

        final Entry new_entry =
          new Entry (entry_group.getBases (), new_embl_entry);

        entry_group.add (new_entry);

        status_line.setText ("");

        last_entry_edit = makeEntryEdit (entry_group);
      }
    }

    for (int entry_index = 0 ;
         entry_index < entry_edit_objects.size () ;
         ++entry_index) {
      entry_edit_objects.elementAt (entry_index).show ();
    }
  }

  /**
   *  Read the entries given in the diana.ini file.
   **/
  private void readDefaultEntries () {
    final String default_sequence_file_name =
      getOptions ().getDefaultSequenceFileName ();

    final String default_feature_file_name =
      getOptions ().getDefaultFeatureFileName ();

    if (default_sequence_file_name != null) {
      final String default_sequence_file_name_embl =
        default_sequence_file_name + "_embl";

      embl.Entry new_embl_entry = null;

      // try opening the default sequence file with "_embl" added to the name
      // if that fails try the plain sequence file name
      if (new File (default_sequence_file_name_embl).exists ()) {
        new_embl_entry = openEntry (default_sequence_file_name_embl);
      }

      if (new_embl_entry == null) {
        new_embl_entry = openEntry (default_sequence_file_name);
      }

      if (new_embl_entry == null) {
        // read failed
        return;
      }

      status_line.setText ("");

      final Bases embl_bases = new Bases (new_embl_entry.getSequence ());

      final EntryGroup entry_group = new EntryGroup (embl_bases);

      final Entry new_entry =
        new Entry (entry_group.getBases (), new_embl_entry);

      entry_group.add (new_entry);

      final EntryEdit new_entry_edit = makeEntryEdit (entry_group);

      if (default_feature_file_name != null) {
        final embl.Entry new_embl_table_entry =
          openEntry (default_feature_file_name);

        if (new_embl_table_entry == null) {
          // open failed
          return;
        }
          
        final Entry new_table_entry =
          new Entry (entry_group.getBases (), new_embl_table_entry);

        entry_group.add (new_table_entry);
      }
    }
  }

  /**
   *  Open the Entry with the given file name or show a message if there is no
   *  document with that name.
   *  @param new_entry_name The file name to open.
   *  @return The reference of the new embl.Entry or null if the read failed.
   **/
  private embl.Entry openEntry (final String new_entry_name) {
    try {
      System.out.println ("trying to open: " + new_entry_name);

      final Document new_document =
        new FileProgressDocument (new File (new_entry_name), status_line);

//      final Entry new_entry = Entry.openEntry (entry_group, new_document);

      if (! new_document.readable ()) {
        // need to create a new Document before continuing
        new MessageDialog (this, "cannot open this file: "  + new_entry_name);
        return null;
      }

      final embl.Entry new_entry = new embl.DocumentEntry (new_document);

//    System.out.println ("Entry.openEntry (): " + entry);

      return new_entry;
    } catch (embl.ReadFormatException e) {
      final String message =
        "failed to open " + new_entry_name + " - " +
        e.getMessage () + " at line: " + e.getLineNumber ();
      System.out.println (message);
      new MessageDialog (this, message);
    } catch (IOException e) {
      final String message =
        "failed to open: " + e.getMessage ();
      status_line.setText (message);
      new MessageDialog (this, message);
    } catch (embl.InvalidRelationException e) {
      final String message =
        "illegal qualifier: " + e.getMessage ();
      status_line.setText (message);
      new MessageDialog (this, message);
    }

    return null;
  }

  /**
   *  The constructor creates all the components for the main Diana window and
   *  sets up all the menu callbacks.
   */
  private Diana () {
    super ("P.O.W.M.A.P.");

    addWindowListener (new WindowAdapter () {
      public void windowClosing (WindowEvent event) {
        exit ();
      }
    });

    setLayout(new BorderLayout());

    final Font font = getOptions ().getFont ();

    setFont (font);

    makeAllMenus ();

    final Canvas helix_canvas = makeHelixCanvas ();

    add (helix_canvas, "Center");

    add (status_line, "South");

    pack ();

    setSize (450, 200);

    final Dimension screen = Toolkit.getDefaultToolkit().getScreenSize();
    setLocation (new Point ((screen.width - getSize ().width) / 2,
                            (screen.height - getSize ().height) / 2));
  }

  /**
   *  This method will fill the background with white.
   **/
  public void paint (final Graphics g) {
    g.setColor (Color.white);
    g.fillRect (0, 0, getSize ().width, getSize ().height);
  }

  /**
   *  Return a Canvas object that will display a helix and a short copyright
   *  notice.
   **/
  private Canvas makeHelixCanvas () {
    return new Canvas () {
      public void update(Graphics g) {
        paint(g);
      }

      public void paint (final Graphics g) {
        if (helix == null) {
          Toolkit toolkit = Toolkit.getDefaultToolkit ();

          final String install_path =
            System.getProperties ().getProperty ("diana.install.path");

          helix = toolkit.getImage (install_path + File.separatorChar +
                                    "diana/helix.gif");

          tracker = new MediaTracker (this);

          tracker.addImage (helix, 0);

          try {
            tracker.waitForAll ();
            helix_height = helix.getHeight (this);
            helix_width = helix.getWidth (this);
          } catch (InterruptedException e) {
            return;
          }
        }

        g.setColor (Color.white);
        g.fillRect (0, 0, getSize ().width, getSize ().height);

        for (int i = 0 ; i * helix_height <= getSize ().height; ++i) {
          g.drawImage (helix,
                       0,
                       i * helix_height, this);
        }

        final Font canvas_font = new Font ("Monospaced", Font.PLAIN, 16);

        g.setFont (canvas_font);

        FontMetrics fm = getFontMetrics (canvas_font);

        final int font_height = fm.getHeight () + 3;

        g.setColor (Color.black);


        g.drawString ("Artemis", helix_width + 5, font_height);
        g.drawString ("Copyright 1998,1999", helix_width + 5, font_height * 2);
        g.drawString ("Genome Research Limited",
                      helix_width + 5, font_height * 3);
      }

      Image helix = null;

      MediaTracker tracker = null;

      int helix_height;
      int helix_width;
    };
  }

  /**
   *  This method gets rid of an EntryEdit object and it's frame.  Each object
   *  removed with this method must have been added previously with
   *  addEntryEdit ().
   *  @param entry_edit The object to get rid of.
   **/
  static public synchronized void entryEditFinished (EntryEdit entry_edit) {
    if (null == entry_edit) {
      throw new Error ("Diana.entryEditFinished () was passed a null object");
    }

    if (!entry_edit_objects.removeElement (entry_edit)) {
      throw new Error ("Diana.entryEditFinished () - could not remove a " +
                       "object from an empty vector");
    }
    entry_edit.setVisible (false);
    entry_edit.dispose ();
  }

  /**
   *  Add an EntryEdit object to our list of objects.
   *  @param entry_edit The object to add.
   **/
  static public synchronized void addEntryEdit (EntryEdit entry_edit) {
    entry_edit_objects.addElement (entry_edit);
  }

  /**
   *  Return the reference of the global clipboard object for this program.
   **/
  static public ClipBoard getClipboard () {
    return clipboard;
  }

  /**
   *  Return the reference of the global options object for this program.
   **/
  static public Options getOptions () {
    if (options == null) {
      options = new Options ();

      try {
        final InputStream options_input_stream =
          getCodeDirectory ().append ("options").getInputStream ();

        options.load (options_input_stream);

        if (isStandAlone ()) {
          final Document current_directory_options_document =
            new FileDocument (new File ("options"));

          // read the options file from the current directory if it is there
          if (current_directory_options_document.readable ()) {

            final InputStream current_directory_options_stream =
              current_directory_options_document.getInputStream ();

            options.load (current_directory_options_stream);
          }

          final Document current_directory_dot_ini_document =
            new FileDocument (new File ("diana.ini"));

          // read the "diana.ini" file from the current directory if it is
          // there
          if (current_directory_dot_ini_document.readable ()) {

            final InputStream current_directory_dot_ini_stream =
              current_directory_dot_ini_document.getInputStream ();

            options.load (current_directory_dot_ini_stream);
          }
        }
      } catch (IOException e) {
        System.out.println ("could not read the options file - " + e);
// XXX
//          new MessageDialog (this,
//                             "could not read the options file - " +
//                             e.getMessage ());
      }
    }
    return options;
  }

  /**
   *  Return true if and only if we are running in an applet.
   **/
  public static boolean isApplet () {
    return is_applet;
  }

  /**
   *  Set whether we are running as an applet.
   **/
  static void setIsApplet (boolean is_applet) {
    Diana.is_applet = is_applet;
  }

  /**
   *  Return true if and only if we are not running in an applet.
   **/
  public static boolean isStandAlone () {
    if (isApplet ()) {
      return false;
    } else {
      return true;
    }
  }

  /**
   *  Return a Document that refers to the directory which contains the code
   *  we are running.
   **/
  public static Document getCodeDirectory () {
    return code_directory;
  }

  /**
   *  Set the Document that refers to the directory which contains the code we
   *  are running.  This will be called from the Diana class constructor or
   *  the DianaApplet constructor.
   **/
  static void setCodeDirectory (Document code_directory) {
    Diana.code_directory = code_directory;
  }

  /**
   *  Make a new menu item in the given menu, with its label given the
   *  String and add the given ActionListener to it.
   */
  private static void makeMenuItem (Menu menu, String name,
                                    ActionListener listener) {
    MenuItem new_item = new MenuItem (name);
    menu.add (new_item);
    new_item.addActionListener (listener);
  }

  /**
   *  Make all the menus and menu items for the main window.  Also sets up
   *  suitable ActionListener objects for each item.
   */
  private void makeAllMenus () {

    setMenuBar (menu_bar);

    menu_bar.add (file_menu);
//    menu_bar.add (edit_menu);
    menu_bar.add (options_menu);

    ActionListener menu_listener;

    menu_listener = new ActionListener () {
      public void actionPerformed (ActionEvent event) {
        makeOpenDialog ();
      }
    };
    makeMenuItem (file_menu, "Open ...", menu_listener);

    menu_listener = new ActionListener () {
      public void actionPerformed (ActionEvent event) {
        makeCorbaDialog ();
      }
    };
    makeMenuItem (file_menu, "Read from EBI ...", menu_listener);

    menu_listener = new ActionListener () {
      public void actionPerformed (ActionEvent event) {
        exit ();
      }
    };
    makeMenuItem (file_menu, "Quit", menu_listener);

    menu_listener = new ActionListener () {
      public void actionPerformed (ActionEvent event) {
        resetOptions ();
      }
    };
    makeMenuItem (options_menu, "Re-read Options", menu_listener);

    final CheckboxMenuItem enable_direct_edit_item =
      new CheckboxMenuItem ("Enable Direct Editing");
    enable_direct_edit_item.setState (canDirectEdit ());
    enable_direct_edit_item.addItemListener (new ItemListener () {
      public void itemStateChanged(ItemEvent event) {
        setDirectEdit (enable_direct_edit_item.getState ());
      }
    });
    options_menu.add (enable_direct_edit_item);
   }

   /**
    *  Force the options files to be re-read and the EntryEdit components to be
    *  redisplayed.
    **/
   private void resetOptions () {
     Diana.options = null;
     // redisplay all the EntryEdit objects
     redisplayAll ();
   }

  /**
   *  Enable direct editing if and only if the argument is true.  "Direct
   *  Editing" allows feature locations to be changed by dragging the ends of
   *  the features around with the mouse.
   **/
  static private void setDirectEdit (final boolean direct_edit) {
    Diana.direct_edit = direct_edit;
  }

  /**
   *  Returns true if and only if direct editing is enabled.
   **/
  static public boolean canDirectEdit () {
    return direct_edit;
  }

  /**
   *  Force all the EntryEdit components to be redisplayed.
   **/
  private void redisplayAll () {
    for (int i = 0 ; i < entry_edit_objects.size () ; ++i) {
      entry_edit_objects.elementAt (i).redisplay ();
    }
  }

  /**
   *  Close the main Diana frame and all EntryEdit frames and then this frame,
   *  then exit.
   **/
  private void exit () {
    for (int i = 0 ; i < entry_edit_objects.size () ; ++i) {
      entryEditFinished (entry_edit_objects.elementAt (i));
    }
    setVisible (false);
    dispose ();
    System.exit (0);
  }

  /**
   *  Create a file dialog and if the user selects a file then call
   *  makeEntryEdit () with the entry.
   **/
  private EntryEdit makeOpenDialog () {
    final EntryFileDialog dialog = new EntryFileDialog (main_window);

    final embl.Entry new_embl_entry = dialog.getEntry ();

    if (new_embl_entry != null) {

      status_line.setText ("");

      final Bases embl_bases = new Bases (new_embl_entry.getSequence ());

      final EntryGroup entry_group = new EntryGroup (embl_bases);

      entry_group.add (new Entry (entry_group.getBases (), new_embl_entry));

      final EntryEdit entry_edit = makeEntryEdit (entry_group);

      entry_edit.setVisible (true);

      return entry_edit;
    }

    return null;
  }

  /**
   *  Create a TextRequester, wait for the user to type an accession number
   *  and then read that entry from the server.
   **/
  private void makeCorbaDialog () {
    final org.omg.CORBA.ORB orb =
      org.omg.CORBA.ORB.init (new String [0], new java.util.Properties());

    final org.omg.CORBA.Object obj;

    try {
      obj = orb.string_to_object (getIOR ());
    } catch (IOException e) {
      final String message =
        "Error while reading EMBL IOR: " + e;
      status_line.setText (message);
      new MessageDialog (this, message);
      return;
    }

    final nsdb.Embl corba_embl = nsdb.EmblHelper.narrow (obj);

    if (corba_embl == null) {
      final String message = "Failed to get an object from Corba/EMBL";
      status_line.setText (message);
      new MessageDialog (this, message);
      return;
    }

    final TextRequester text_requester =
      new TextRequester ("Enter an accession number:", 10, "");

    final TextRequesterListener listener =
      new TextRequesterListener () {
        public void actionPerformed (final TextRequesterEvent event) {
          if (event.getType () == TextRequesterEvent.CANCEL) {
            return;
          }

          final String accession_number = event.getRequesterText ().trim ();

          makeFromAccessionNumber (corba_embl, accession_number);
        }
      };

    text_requester.addTextRequesterListener (listener);

    text_requester.setVisible (true);
  }

  private void makeFromAccessionNumber (final nsdb.Embl corba_handle,
                                        final String accession_number) {
    try {
      final EmblSeq embl_seq = corba_handle.getEmblSeq (accession_number);
      final embl.Entry new_embl_entry = new embl.CorbaEntry (embl_seq);

      final Bases embl_bases = new Bases (new_embl_entry.getSequence ());

      final EntryGroup entry_group = new EntryGroup (embl_bases);

      final Entry new_entry =
        new Entry (entry_group.getBases (), new_embl_entry);

      entry_group.add (new_entry);

      status_line.setText ("");

      final EntryEdit entry_edit = makeEntryEdit (entry_group);

      entry_edit.setVisible (true);
    } catch (NoResult e) {
      final String message =
        "Database query failed while getting accession number: " +
        accession_number;
      status_line.setText (message);
      new MessageDialog (Diana.this, message);
    } catch (Superceded e) {
      //  Superceded is used by getEmblSeq method if accession number
      //  doesn't exist anymore because it was merged or split
      final String message =
        "This accession number has been superceded: " + accession_number;
      status_line.setText (message);
      new MessageDialog (Diana.this, message);
    } catch (embl.LocationParseException e) {
      final String message =
        "Unexpected error while accessing " + accession_number + ": " + e;
      status_line.setText (message);
      new MessageDialog (Diana.this, message);
    } catch (embl.InvalidKeyException e) {
      final String message =
        "Unexpected error while accessing " + accession_number + ": " + e;
      status_line.setText (message);
      new MessageDialog (Diana.this, message);
    } catch (java.lang.Throwable e) {
      final String message =
        "Unexpected error while accessing " + accession_number + ": " + e;
      status_line.setText (message);
      new MessageDialog (Diana.this, message);

      e.printStackTrace ();
    }
  }

  /**
   *  Finds and returns a stringified object reference (IOR).
   *  @return The stringified IOR.
   *  @exception IOException Thrown if an error occurs while reading.
   **/
  private String getIOR ()
      throws IOException {
    try {
      //  get URL of IOR
      URL url = new URL (default_ior_url);
      BufferedReader in =
        new BufferedReader (new InputStreamReader (url.openStream()));
      return in.readLine();
    } catch (MalformedURLException e) {
      throw new Error ("internal error - unexpected MalformedURLException: " +
                       e.getMessage ());
    }
  }

  /**
   *  Create a EntryEdit object from an EMBL file and then call
   *  addEntryEdit ().
   *  @param entry The embl.Entry object for the new EntryEdit object.
   *  @return The newly created EntryEdit component.
   **/
  public static EntryEdit makeEntryEdit (EntryGroup entry_group) {

    final EntryEdit entry_edit = new EntryEdit (entry_group);

    addEntryEdit (entry_edit);

    return entry_edit;
  }

  /**
   *  The default place to look for the IOR for the corba server.
   **/
  static private final String default_ior_url =
    "http://corba.ebi.ac.uk/EMBL/IOR/Embl.IOR";

  /**
   *  The main diana frame.
   **/
  static private Diana main_window;

  /**
   *  Menu bar for the main window.
   **/
  static final private MenuBar menu_bar = new MenuBar ();

  /**
   *  The file menu.
   **/
  static final private Menu file_menu = new Menu ("File");

  /**
   *   The edit menu.
   **/
  static final private Menu edit_menu = new Menu ("Edit");

  /**
   *   The options menu.
   **/
  static final private Menu options_menu = new Menu ("Options");

  /**
   *
   **/
  static final private Panel panel = new Panel ();

  /**
   *  A label for status and error messages.
   **/
  static final private Label status_line = new Label ("");

  /**
   *  A vector containing all the EntryEdit object we have created.
   **/
  static private EntryEditVector entry_edit_objects = new EntryEditVector ();

  /**
   *  The global clip board object.  Used by the stand alone program and the
   *  applet.
   **/
  static private ClipBoard clipboard = new ClipBoard ();

  /**
   *  The global options object.  Used by the stand alone program and the
   *  applet.
   **/
  static private Options options = null;

  /**
   *  This is true if and only if we are running as an applet.
   **/
  static private boolean is_applet = false;

  /**
   *  This contains a Document that refers to the directory which contains the
   *  code we are running.
   **/
  static private Document code_directory = null;

  /**
   *  Direct editing is enabled if and only if this is true.  "Direct editing"
   *  allows feature locations to be changed by dragging the ends of features
   *  around with the mouse.
   **/
  static private boolean direct_edit = false;
}


