/* EntryGroupInfoDisplay.java
 *
 * created: Fri Mar 12 1999
 *
 * This file is part of Artemis
 * 
 * Copyright (C) 1998,1999  Genome Research Limited
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * $Header: /nfs/disk222/yeastpub/Repository/powmap/diana/components/EntryGroupInfoDisplay.java,v 1.3 1999/04/16 12:40:06 kmr Exp $
 **/

package diana.components;

import diana.*;
import diana.sequence.*;

import java.util.Hashtable;
import java.util.Enumeration;
import java.awt.event.*;

/**
 *  This component will show general information about an EntryGroup.  It will
 *  show the sequence length, GC content and a summary of the active entries.
 *
 *  @author Kim Rutherford
 *  @version $Id: EntryGroupInfoDisplay.java,v 1.3 1999/04/16 12:40:06 kmr Exp $
 **/

public class EntryGroupInfoDisplay
    implements FeatureChangeListener, EntryChangeListener,
               EntryGroupChangeListener, SequenceChangeListener {
  /**
   *  Create a new EntryGroupInfoDisplay object to summarize the given
   *  EntryGroup.
   **/
  public EntryGroupInfoDisplay (final EntryGroup entry_group) {
    file_viewer = new FileViewer ("P.O.W.M.A.P Overview");

    this.entry_group = entry_group;
    
    updateView ();

    entry_group.addEntryChangeListener (this);
    entry_group.addFeatureChangeListener (this);
    entry_group.addEntryGroupChangeListener (this);
    entry_group.getBases ().addSequenceChangeListener (this);

    file_viewer.close_button.addActionListener (new ActionListener () {
      public void actionPerformed (ActionEvent e) {
        stopListening ();
      }
    });
    
    file_viewer.addWindowListener (new WindowAdapter () {
      public void windowClosing (WindowEvent event) {
        stopListening ();
      }
    });
  }

  /**
   *  Remove this object as a selection change listener.
   **/
  private void stopListening () {
    entry_group.removeEntryChangeListener (this);
    entry_group.removeEntryGroupChangeListener (this);
    entry_group.removeFeatureChangeListener (this);
    entry_group.getBases ().removeSequenceChangeListener (this);
  }
  
  /**
   *  Implementation of the FeatureChangeListener interface.   We listen to
   *  FeatureChange events so that we can update the display if qualifiers
   *  change.
   **/
  public void featureChanged (final FeatureChangeEvent event) {
    updateView ();
  }

  /**
   *  Implementation of the EntryGroupChangeListener interface.  We listen to
   *  EntryGroupChange events so that we can update the display if entries
   *  are added, deleted, activated or deactivated.
   **/
  public void entryGroupChanged (final EntryGroupChangeEvent event) {
    updateView ();
  }

  /**
   *  Implementation of the EntryChangeListener interface.  We listen to
   *  EntryChange events so that we can update the display if features are
   *  added or deleted.
   **/
  public void entryChanged (final EntryChangeEvent event) {
    updateView ();
  }

  /**
   *  Implementation of the SequenceChangeListener interface.  We listen to
   *  SequenceChange events so that we can update the display if the sequence
   *  changes.
   **/
  public void sequenceChanged (final SequenceChangeEvent event) {
    updateView ();
  }

  /**
   *  Update the FileViewer component to reflex the current state of the
   *  EntryGroup.
   **/
  private void updateView () {
    final StringBuffer buffer = new StringBuffer ();

    buffer.append ("Number of bases: " +
                   entry_group.getSequenceLength () + "\n");
    buffer.append ("Number of features in the active entries: " +
                   entry_group.getAllFeaturesCount () + "\n");
    buffer.append ("Percentage GC in sequence: " +
                   entry_group.getBases ().getAverageGCPercent () + "\n");

    final Hashtable table = new Hashtable ();

    final FeatureEnumeration feature_enumerator = entry_group.features ();

    while (feature_enumerator.hasMoreFeatures ()) {
      final Feature this_feature = feature_enumerator.nextFeature ();

      final String key_string = this_feature.getKey ().toString ();

      if (table.containsKey (key_string)) {
        final int old_value = ((Integer) table.get (key_string)).intValue ();
  
        table.put (key_string, new Integer (old_value + 1));
      } else {
        table.put (key_string, new Integer (1));
      }
    }

    buffer.append ("\nSummary of the active entries:\n");
    
    final Enumeration e = table.keys ();

    while (e.hasMoreElements()) {
      final String this_key = (String) e.nextElement();
      final int this_value = ((Integer) table.get (this_key)).intValue ();

      buffer.append (this_key + ": " + this_value + "\n");
    }

    file_viewer.readString (buffer.toString ());
  }
  
  /**
   *  This is the EntryGroup object that we are viewing.
   **/
  final private EntryGroup entry_group;

  /**
   *  The FileViewer object that is displaying the EntryGroup.
   **/
  final private FileViewer file_viewer;
}


