/* SelectMenu.java
 *
 * created: Thu Jan 14 1999
 *
 * This file is part of Artemis
 *
 * Copyright (C) 1998,1999  Genome Research Limited
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * $Header: /nfs/disk222/yeastpub/Repository/powmap/diana/components/SelectMenu.java,v 1.10 1999/05/05 10:29:26 kmr Exp $
 **/

package diana.components;

import diana.*;
import embl.Key;
import embl.KeyVector;
import embl.Qualifier;
import embl.InvalidKeyException;

import java.awt.*;
import java.awt.event.*;

/**
 *  This menu has contains items such a "Select all", "Select none" and
 *  "Select by type".
 *
 *  @author Kim Rutherford
 *  @version $Id: SelectMenu.java,v 1.10 1999/05/05 10:29:26 kmr Exp $
 **/

public class SelectMenu extends SelectionMenu {
  /**
   *  Create a new SelectMenu object and all it's menu items.
   *  @param frame The Frame that owns this Menu.
   *  @param selection_owner The owner of the selection that the commands in
   *    the menu will operate on.
   *  @param entry_group The EntryGroup object where features and exons are
   *    selected from.
   **/
  public SelectMenu (final Frame frame,
                     final SelectionOwner selection_owner,
                     final EntryGroup entry_group) {
    super (frame, "Select", selection_owner);

    this.entry_group = entry_group;

    all_item = new MenuItem ("Select All",
                             new MenuShortcut (KeyEvent.VK_A));
    all_item.addActionListener (new ActionListener () {
      public void actionPerformed (ActionEvent event) {
        selectAll ();
      }
    });

    add (all_item);

    none_item = new MenuItem ("Select None",
                              new MenuShortcut (KeyEvent.VK_N));
    none_item.addActionListener (new ActionListener () {
      public void actionPerformed (ActionEvent event) {
        getSelectionOwner ().getSelection ().clear ();
      }
    });

    add (none_item);

    select_by_key_item = new MenuItem ("Select By Type");
    select_by_key_item.addActionListener (new ActionListener () {
      public void actionPerformed (ActionEvent event) {
        selectByTypePopup ();
      }
    });

    add (select_by_key_item);

    select_cds_item = new MenuItem ("Select CDS Features",
                                    new MenuShortcut (KeyEvent.VK_F));
    select_cds_item.addActionListener (new ActionListener () {
      public void actionPerformed (ActionEvent event) {
        // select all CDS features that do not have the /pseudo qualifier
        final FeatureKeyQualifierPredicate predicate =
          new FeatureKeyQualifierPredicate (Key.CDS, "pseudo", false);

        getSelection ().clear ();

        selectFeaturesByPredicate (predicate);
      }
    });

    add (select_cds_item);

    select_same_type_item = new MenuItem ("Select Same Type");
    select_same_type_item.addActionListener (new ActionListener () {
      public void actionPerformed (ActionEvent event) {
        selectSameType ();
      }
    });

    add (select_same_type_item);

    addSeparator ();

    selection_toggle_item = new MenuItem ("Toggle Selection",
                                          new MenuShortcut (KeyEvent.VK_T));
    selection_toggle_item.addActionListener (new ActionListener () {
      public void actionPerformed (ActionEvent event) {
        toggleSelection ();
      }
    });

    add (selection_toggle_item);
  }

  /**
   *  Select all the features in the active entries.  If there are more than
   *  1000 features the user is asked for confirmation.
   **/
  private void selectAll () {
    final Selection selection = getSelectionOwner ().getSelection ();
    final FeatureVector all_features = entry_group.getAllFeatures ();

    if (all_features.size () > 1000) {
      final YesNoDialog dialog =
        new YesNoDialog (getParentFrame (),
                         "Are you sure you want to select " +
                         all_features.size () + " features?");

      if (!dialog.getResult ()) {
        return;
      }
    }

    selection.set (all_features);
  }


  /**
   *  Invert the selection - after calling this method the selection will
   *  contain only those features that were not in the selection before the
   *  call.
   **/
  private void toggleSelection () {
    final Selection selection = getSelectionOwner ().getSelection ();
    final FeatureVector selected_features = getSelection ().getAllFeatures ();

    final FeatureVector all_features = entry_group.getAllFeatures ();

    final FeatureVector new_selection = new FeatureVector ();

    for (int i = 0 ; i < all_features.size () ; ++i) {
      final Feature this_feature = all_features.elementAt (i);

      if (!selected_features.contains (this_feature)) {
        new_selection.add (this_feature);
      }
    }

    selection.set (new_selection);
  }
  
  
  /**
   *  Popup a TextRequester component and then select all the features that
   *  have the same key as the user types.
   **/
  private void selectByTypePopup () {
    final TextRequester text_requester =
      new TextRequester ("Select which type?", 18, "");

    text_requester.addTextRequesterListener (new TextRequesterListener () {
      public void actionPerformed (final TextRequesterEvent event) {
        final String requester_text = event.getRequesterText ().trim ();
        if (requester_text.length () == 0) {
          new MessageDialog (getParentFrame (), "key string too short");
        } else {
          final Key key;

          try {
            key = new Key (requester_text);
          } catch (InvalidKeyException e) {
            new MessageDialog (getParentFrame (),
                               "invalid key: " + requester_text);
            return;
          }

          getSelection ().clear ();

          selectByKey (key);
        }
      }
    });

    text_requester.show ();
  }

  /**
   *  Select all the features that have the given key.
   **/
  private void selectByKey (final Key key) {
    final FeatureKeyPredicate predicate = new FeatureKeyPredicate (key);

    selectFeaturesByPredicate (predicate);
  }

  /**
   *  Select all the features that have the same key as the currently selected
   *  features.
   **/
  private void selectSameType () {
    if (! checkForSelectionFeatures ()) {
      return;
    }

    final KeyVector seen_keys = new KeyVector ();

    final FeatureVector selected_features = getSelection ().getAllFeatures ();

    for (int i = 0 ; i < selected_features.size () ; ++i) {
      final Feature current_feature = selected_features.elementAt (i);

      if (!seen_keys.contains (current_feature.getKey ())) {
        seen_keys.add (current_feature.getKey ());
      }
    }

    getSelection ().clear ();

    for (int i = 0 ; i < seen_keys.size () ; ++i) {
      selectByKey (seen_keys.elementAt (i));
    }
  }

  /**
   *  Select all the features that match the given predicate.
   **/
  private void selectFeaturesByPredicate (final FeaturePredicate predicate) {
    final FeatureVector new_selection_features = new FeatureVector ();

    final FeatureEnumeration feature_enum = entry_group.features ();

    while (feature_enum.hasMoreFeatures ()) {
      final Feature current_feature = feature_enum.nextFeature ();

//      System.out.println ("adding a feature: " + new_feature.getKey ());

      if (predicate.testPredicate (current_feature)) {
        new_selection_features.add (current_feature);
      }
    }

    getSelection ().add (new_selection_features);
  }

  /**
   *  The EntryGroup object that was passed to the constructor.
   **/
  private EntryGroup entry_group;

  private MenuItem all_item = null;
  private MenuItem none_item = null;
  private MenuItem selection_toggle_item = null;
  private MenuItem select_same_type_item = null;
  private MenuItem select_by_key_item = null;
  private MenuItem select_cds_item = null;
}
