/* UserFeatureAminoAcidAlgorithm.java
 *
 * created: Mon Dec 21 1998
 *
 * This file is part of Artemis
 * 
 * Copyright (C) 1998,1999  Genome Research Limited
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * $Header: /nfs/disk222/yeastpub/Repository/powmap/diana/plot/UserFeatureAminoAcidAlgorithm.java,v 1.2 1999/03/22 11:00:43 kmr Exp $
 **/

package diana.plot;

import diana.*;
import diana.sequence.*;
import embl.Range;

/**
 *  Objects of this class have one useful method - getValues (), which takes a
 *  range of bases and returns a single floating point number.  The return
 *  value is calculated by comparing the  to the weight matrix that is
 *  passed to constructor.
 *
 *  @author Kim Rutherford
 *  @version $Id: UserFeatureAminoAcidAlgorithm.java,v 1.2 1999/03/22 11:00:43 kmr Exp $
 **/

public class UserFeatureAminoAcidAlgorithm extends FeatureAlgorithm {
  /**
   *  Create a new UserBaseAlgorithm object.  See getValues () for detail on
   *  the weightings array.
   *  @param feature The feature to do future calculation on.
   *  @param algorithm_name The algorithm name.
   *  @param default_window_size The default or optimal window size.
   *  @param min_window_size The minimum window size for this algorithm.
   *  @param max_window_size The maximum window size for this algorithm.
   *  @param geometric_mean If true then the return value(s) of getValues ()
   *    will be calculated using a geometric mean rather than an arithmetic.
   *  @param weightings The weight matrix to use in getValues ().  The first
   *    index of this arry is an index from 0 to 19 over the amino acids in
   *    the order given by AminoAcidSequence.getSymbolIndex ().  The second
   *    index is a residue position in the translation.  All rows of the
   *    matrix must be the same length (and at least 1 long).  All columns
   *    must be 20 in length (one row for each residue).
   **/
  public UserFeatureAminoAcidAlgorithm (Feature feature,
                                        String algorithm_name,
                                        int default_window_size,
                                        int min_window_size,
                                        int max_window_size,
                                        boolean geometric_mean,
                                        float [] [] weightings) {
    super (feature, algorithm_name);
    this.default_window_size = default_window_size;
    this.min_window_size = min_window_size;
    this.max_window_size = max_window_size;
    this.geometric_mean = geometric_mean;
    this.weightings = weightings;
  }

  /**
   *  Return the average value of the user weightings for the sequence of
   *  bases.  The order to list the residue weights is given by
   *  AminoAcidSequence.getSymbolIndex ().
   *  An example - a UserFeatureAminoAcidAlgorithm created like this:
   *  <pre>
   *   float [] [] weights = { {2,3,4,5,...}, {1,1,1,1,...}, {0,2,6,7,...} };
   *   new UserFeatureAminoAcidAlgorithm ("test", weights);
   *  </pre>
   *  with the sequence "rrr" will give (3+1+2)/3 = 2.0, whereas "arn" will
   *  give (2+1+6)/3 = 3.0
   *  @param start The start base (included in the range).
   *  @param end The end base (included in the range).
   *  @param values The one return value for this algorithm is returned in
   *    this array.
   **/
  public void getValues (int start, int end, float [] values) {
    final int number_of_weights = weightings[0].length;

    final AminoAcidSequence translation = getFeature ().getTranslation ();

    final String translation_string =
      translation.toString ().substring (start, end);

    float total;

    if (geometric_mean) {
      total = 1;
    } else {
      total = 0;
    }

    for (int i = 0 ; i < translation_string.length () ; ++i) {
      final char this_char =
        translation_string.charAt (i % translation_string.length ());

      final float this_score = getScore (this_char);

//      System.out.println ("this_score: " + this_score);
      
      if (geometric_mean) {
//          System.out.println (weightings[this_symbol_index][this_weight_index] +
//                              " " + total + " " + this_char);

//        System.out.print (this_char);

        total *= this_score;
      } else {
        total += this_score;
      }
    }

    if (geometric_mean) {
      values[0] = (float) Math.pow (total, 1.0 / translation_string.length ());
//        System.out.println ("value[0]: " + values[0] + " " +
//                            Math.pow (total, 1.0 / number_of_weights) + " " +
//                            total + " " + 1.0 / number_of_weights);
    } else {
      values[0] = total / number_of_weights;
    }
  }

  /**
   *  Return the score for the given residue.
   **/
  private float getScore (char this_residue) {
    final int number_of_weights = weightings[0].length;

    float max_score = Float.MIN_VALUE;

    for (int weight_index = 0 ;
         weight_index < number_of_weights ;
         ++weight_index) {

      final int this_weight_index = weight_index % number_of_weights;
      final int this_symbol_index =
        AminoAcidSequence.getSymbolIndex (this_residue);

//          System.out.println (weightings[this_symbol_index][this_weight_index] +
//                              " " + total + " " + this_char);

//        System.out.print (this_residue);

        if (weightings[this_symbol_index][weight_index] > max_score) {
          max_score = weightings[this_symbol_index][weight_index];
        }
    }

    return max_score;
  }

  /**
   *  Return the score for the given sequence of residues in the given frame.
   **/
  private float getScore (String translation_string, int frame) {
    final int number_of_weights = weightings[0].length;

    float total;

    if (geometric_mean) {
      total = 1;
    } else {
      total = 0;
    }

    for (int i = 0 ; i < translation_string.length () ; ++i) {
      final char this_char =
        translation_string.charAt ((i + frame) % translation_string.length ());

      final int this_weight_index = i % number_of_weights;
      final int this_symbol_index =
        AminoAcidSequence.getSymbolIndex (this_char);

      if (geometric_mean) {
//          System.out.println (weightings[this_symbol_index][this_weight_index] +
//                              " " + total + " " + this_char);

        System.out.print (this_char);

        total *= weightings[this_symbol_index][this_weight_index];
      } else {
        total += weightings[this_symbol_index][this_weight_index];
      }
    }


//    System.out.println ("total: " + total + " " + sequence);

    if (geometric_mean) {
      return (float) Math.pow (total, 1.0 / translation_string.length ());
//        System.out.println ("value[0]: " + values[0] + " " +
//                            Math.pow (total, 1.0 / number_of_weights) + " " +
//                            total + " " + 1.0 / number_of_weights);
    } else {
      return total / number_of_weights;
    }
  }

  /**
   *  Return the number of values a call to getValues () will return - one
   *  in this case.
   **/
  public int getValueCount () {
    return 1;
  }

  /**
   *  Return the default or optimal window size.
   *  @return null is returned if this algorithm doesn't have optimal window
   *    size.
   **/
  public Integer getDefaultWindowSize () {
    return new Integer (default_window_size);
  }

  /**
   *  Return the default maximum window size for this algorithm.
   *  @return null is returned if this algorithm doesn't have maximum window
   *    size.
   **/
  public Integer getDefaultMaxWindowSize () {
    return new Integer (max_window_size);
  }

  /**
   *  Return the default minimum window size for this algorithm.
   *  @return null is returned if this algorithm doesn't have minimum window
   *    size.
   **/
  public Integer getDefaultMinWindowSize () {
    return new Integer (min_window_size);
  }

  /**
   *  Return the default or optimal step size.
   *  @return null is returned if this algorithm doesn't have optimal step
   *    size.
   **/
  public Integer getDefaultStepSize (int window_size) {
    return new Integer (1);
  }

  /**
   *  Return the average value of function over the whole strand.
   *  @return null is returned if this algorithm doesn't have an average or if
   *    the average can't be calculated.
   **/
  public Float getAverage () {
    return null;
  }

  /**
   *  The weightings array that was passed to the constructor.
   **/
  private float [] [] weightings = null;

  /**
   *   The default or optimal window size, as passed to the constructor.
   **/
  private int default_window_size = 0;

  /**
   *  The minimum window size for this algorithm, as passed to the constructor.
   **/
  private int min_window_size = 0;

  /**
   *  The maximum window size for this algorithm, as passed to the constructor.
   **/
  private int max_window_size = 0;

  /**
   *  If true then the return value(s) of getValues () will be calculated
   *  using a geometric mean rather than an arithmetic, as passed to the
   *  constructor.
   **/
  private boolean geometric_mean = false;
}


