/* EntryInformation.java
 *
 * created: Tue Jan  5 1999
 *
 * This file is part of Artemis
 * 
 * Copyright (C) 1998,1999  Genome Research Limited
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * $Header: /nfs/disk222/yeastpub/Repository/powmap/embl/EntryInformation.java,v 1.8 1999/03/22 11:00:55 kmr Exp $
 **/

package embl;

import Document;
import StringVector;

import java.util.*;
import java.io.*;

/**
 *  This class contains static methods that give information about the format
 *  of an EMBL entry, such as the possible keys and qualifiers.  It must be
 *  initialised with a call to initialise () before any other methods can be
 *  accessed.
 *
 *  @author Kim Rutherford
 *  @version $Id: EntryInformation.java,v 1.8 1999/03/22 11:00:55 kmr Exp $
 **/

public class EntryInformation {
  /**
   *  This method must be called before any other methods can be accessed.  It
   *  will read the possible feature key type and qualifiers name and types
   *  from the files qualifier_types and feature_keys.
   *  @param root_directory This object must give the directory which contains
   *    the qualifier_types and and feature_keys files.
   *  @param extra_keys This a list of non-embl keys that should be included
   *    in the return value of getPossibleKeys (), getSortedPossibleKeys ().
   *    This may be null if there are no extra keys.
   *  @param extra_qualifiers This a list of non-embl qualifiers that should be
   *    included in the return value of  (),
   *    getPossibleQualifierNames ().
   *    This may be null if there are no extra qualifiers.
   *  @exception IOException Thrown if either of the required files can't be
   *    opened or if is an input/output error while reading from them.
   *  @exception QualifierInfoException Thrown if the qualifier_types file
   *    doesn't make sense.
p   **/
  public static void initialise (final Document root_directory,
                                 final StringVector extra_keys,
                                 final QualifierInfoVector extra_qualifiers)
      throws IOException, QualifierInfoException {

    if (feature_properties != null) {
      throw new Error ("initernal error - " +
                       "the embl package has been initialised twice");
    }

    EntryInformation.extra_qualifiers = extra_qualifiers;
    EntryInformation.extra_keys = extra_keys;

    if (extra_qualifiers == null) {
      EntryInformation.extra_qualifiers = new QualifierInfoVector ();
    }

    if (extra_keys == null) {
      EntryInformation.extra_keys = new StringVector ();
    }

    final Document feature_keys_document =
      root_directory.append ("feature_keys");

    final InputStream keys_read_stream =
      feature_keys_document.getInputStream ();

    feature_properties = new Properties ();

    feature_properties.load (keys_read_stream);

    initialiseQualifiers (root_directory);

    keys_read_stream.close ();
  }

  /**
   *  Initialise qualifier_properties from the qualifier_types file.
   *  @param root_directory This object must give the directory which contains
   *    the qualifier_types file.
   **/
  private static void initialiseQualifiers (final Document root_directory)
      throws IOException, QualifierInfoException {
    final Document qualifier_document =
      root_directory.append ("qualifier_types");

    final InputStream qualifier_read_stream =
      qualifier_document.getInputStream ();

    final Properties qualifier_properties = new Properties ();

    qualifier_properties.load (qualifier_read_stream);

    final Enumeration qualifier_keys = qualifier_properties.propertyNames ();

    while (qualifier_keys.hasMoreElements ()) {

      final String current_key = (String) qualifier_keys.nextElement ();

      final StringVector current_data_values =
        getPropertyValues (qualifier_properties, current_key);

      final String type_string = current_data_values.elementAt (0);

      qualifier_info_vector.add (new QualifierInfo (current_key,
                                                    type_string));
    }
  }

  /**
   *  Add a new feature key to the list of keys returned by getPossibleKeys ()
   *  and getSortedPossibleKeys ().
   **/
  public static void addPossibleKey (String key_string) {
    extra_keys.add (key_string);
  }

  /**
   *  Return a alphanumerically sorted vector containing the String
   *  representations of the possible keys.
   **/
  public static StringVector getSortedPossibleKeys () {
    final StringVector return_vector = getPossibleKeys ();
    return_vector.sort ();
    return return_vector;
  }

  /**
   *  Return a vector containing the possible feature keys.  The returned
   *  StringVector is a copy.
   *  @see #initialise(Document, StringVector, StringVector)
   **/
  public static StringVector getPossibleKeys () {
    final StringVector return_strings = getPossibleEMBLKeys ();

    for (int i = 0 ; i < extra_keys.size () ; ++i) {
      return_strings.add (extra_keys.elementAt (i));
    }

    return return_strings;
  }

  /**
   *  Return a vector containing the possible embl feature keys.  The vector
   *  will contain only those keys that are legal in an EMBL entry.  The
   *  returned StringVector is a copy.
   *  @see #initialise(Document, StringVector, StringVector)
   **/
  public static StringVector getPossibleEMBLKeys () {
    if (feature_properties == null) {
      throw new Error ("initernal error - " +
                       "the embl package has not been initialised");
    }

    final StringVector return_strings = new StringVector ();

    final Enumeration all_keys = feature_properties.propertyNames ();

    while (all_keys.hasMoreElements ()) {
      final String current_key = (String) all_keys.nextElement ();

      return_strings.add (current_key);
    }

    return return_strings;
  }

  /**
   *  Return a vector containing all possible feature qualifiers.  The
   *  returned StringVector is a copy.
   **/
  public static StringVector getPossibleQualifierNames () {
    final StringVector return_strings = new StringVector ();

    for (int i = 0 ; i < qualifier_info_vector.size () ; ++i) {
      final QualifierInfo current_qualifier_info =
        (QualifierInfo) qualifier_info_vector.elementAt (i);

      return_strings.add (current_qualifier_info.getName ());
    }

    for (int i = 0 ; i < extra_qualifiers.size () ; ++i) {
      final QualifierInfo current_qualifier_info =
        (QualifierInfo) extra_qualifiers.elementAt (i);

      return_strings.add (current_qualifier_info.getName ());
    }

    return return_strings;
  }

  /**
   *  Return a vector containing all possible qualifiers for a feature with
   *  the given key.  The returned StringVector is a copy.
   **/
  public static StringVector getPossibleQualifierNames (final Key key) {
    final StringVector return_strings = new StringVector ();

    return_strings.add (getPropertyValues (feature_properties,
                                           key.toString ()));

    for (int i = 0 ; i < extra_qualifiers.size () ; ++i) {
      final QualifierInfo current_qualifier_info =
        (QualifierInfo) extra_qualifiers.elementAt (i);

      return_strings.add (current_qualifier_info.getName ());
    }

    return return_strings;
  }

  /**
   *  Return true if and only if the given String contains a possible
   *  qualifier name.
   **/
  public static boolean isPossibleQualifier (final String name) {
    if (isPossibleEMBLQualifier (name)) {
      return true;
    } else {
      return isNonStandardQualifier (name);
    }
  }

  /**
   *  Return true if and only if the given String contains a valid EMBL
   *  qualifier name.
   **/
  public static boolean isPossibleEMBLQualifier (final String name) {
    for (int i = 0 ; i < qualifier_info_vector.size () ; ++i) {
      final QualifierInfo current_qualifier_info =
        (QualifierInfo) qualifier_info_vector.elementAt (i);

      if (name.equals (current_qualifier_info.getName ())) {
        return true;
      }
    }

    return false;
 }

  /**
   *  Return true if and only if the given String contains a legal qualifier
   *  name that isn't a valid EMBL qualifier (meaning it is in the
   *  extra_qualifiers property of the options file).
   **/
  public static boolean isNonStandardQualifier (final String name) {
    for (int i = 0 ; i < extra_qualifiers.size () ; ++i) {
      final QualifierInfo current_qualifier_info =
        (QualifierInfo) extra_qualifiers.elementAt (i);

      if (name.equals (current_qualifier_info.getName ())) {
        return true;
      }
    }

    return false;
  }



  /**
   *  Return true if and only if given qualifier_name is a legal qualifier
   *  name for the given key.
   **/
  public static boolean isPossibleQualifier (final Key key,
                                             final String qualifier_name) {
    final StringVector values =
      getPropertyValues (feature_properties, key.toString ());

    if (values == null) {
      // the key is not a standard embl key, so just check that qualifier_name
      // is a legal qualifier
      return isPossibleQualifier (qualifier_name);
    } else {
      if (values.contains (qualifier_name) ||
          values.contains ('@' + qualifier_name)) {
        return true;
      } else {
        if (isNonStandardQualifier (qualifier_name)) {
          return true;
        } else {
          return false;
        }
      }
    }
  }

  /**
   *  Return a String reference containing the qualifier_name if
   *  qualifier_name is a valid qualifier name, otherwise returns null.  The
   *  returned reference refers to a single object, so, for example, passing
   *  any String containing "gene" will always return the same reference (to a
   *  String object containing "gene", whose reference is stored in this
   *  package).
   **/
  public static String checkQualifier (final String qualifier_name) {
    final QualifierInfo qualifier_info = getQualifierInfo (qualifier_name);

    if (qualifier_info == null) {
      return null;
    } else {
      return qualifier_info.getName ();
    }
  }

  /**
   *  Return true if and only if given key_string is a valid Key.  Valid keys
   *  are those that valid in EMBL entries and those that are given in the
   *  extra_keys option.
   **/
  public static boolean isPossibleKey (final String key_string) {
    if (isPossibleEMBLKey (key_string)) {
      return true;
    } else {
      return isNonStandardKey (key_string);
    }
  }

  /**
   *  Return true if and only if given key_string is valid as an EMBL key.
   **/
  public static boolean isPossibleEMBLKey (final String key_string) {
    if (feature_properties == null) {
      throw new Error ("initernal error - " +
                       "the embl package has not been initialised");
    }

    if (feature_properties.getProperty (key_string) != null ||
        feature_properties.getProperty ('@' + key_string) != null) {
      return true;
    } else {
      return false;
    }
  }

  /**
   *  Return true if and only if given key_string is a legal Key, but is not a
   *  standard embl key.  These non standard keys are set by reading the
   *  extra_keys option.
   **/
  public static boolean isNonStandardKey (final String key_string) {
    if (extra_keys.contains (key_string)) {
      return true;
    } else {
      return false;
    }
  }

  /**
   *  Return the QualifierInfo object for the given qualifier name, or null if
   *  there are no legal qualifiers with the given name.
   **/
  public static QualifierInfo getQualifierInfo (final String qualifier_name) {
    for (int i = 0 ; i < qualifier_info_vector.size () ; ++i) {
      final QualifierInfo current_qualifier_info =
        (QualifierInfo) qualifier_info_vector.elementAt (i);

      if (qualifier_name.equals (current_qualifier_info.getName ())) {
        return current_qualifier_info;
      }
    }

    for (int i = 0 ; i < extra_qualifiers.size () ; ++i) {
      final QualifierInfo current_qualifier_info =
        (QualifierInfo) extra_qualifiers.elementAt (i);

      if (qualifier_name.equals (current_qualifier_info.getName ())) {
        return current_qualifier_info;
      }
    }

    return null;
  }

  /**
   *  Return a StringVector containing the values of the given property in the
   *  given Properties object or null if there is no such property.  The
   *  values are separated by spaces in the properties file.  For example,
   *  asking for the property foo in a Properties object where foo has the
   *  value: "thing1 thing2 thing3", will return a vector containing three
   *  String objects: "thing1", "thing2", "thing3"
   **/
  public static StringVector getPropertyValues (final Properties properties,
                                                final String property) {
    final String value = properties.getProperty (property);

    if (value == null) {
      return null;
    }

    return StringVector.getStrings (value);
  }

  /**
   *  This is initialised by the initialise () method to contain descriptions
   *  of the format of each qualifier.  Each object in the Vector will be a
   *  QualifierInfo obect.
   **/
  private static QualifierInfoVector qualifier_info_vector =
    new QualifierInfoVector ();

  /**
   *  This is initialised by the initialise () method to contain the possible
   *  feature keys and the allowable qualifiers for each key.  Each property
   *  in this object contains the feature key as it's key and a String
   *  containing the qualifiers as the data for that key.
   **/
  private static Properties feature_properties = null;

  /**
   *  This a list of non-embl keys that should be included in the return value
   *  of getPossibleKeys (), getSortedPossibleKeys ().
   *  This may be null if there are no extra keys.
   **/
  private static StringVector extra_keys = null;

  /**
   *  This a list of non-embl qualifies that should be included in the return
   *  value of getQualifiersFromKey (), getSortedQualifiersFromKey ().
   *  Each element of the Vector is a QualifierInfo object.
   **/
  private static QualifierInfoVector extra_qualifiers = null;
}
