/* StreamQualifier.java
 *
 * created: Wed Dec 30 1998
 *
 * This file is part of Artemis
 * 
 * Copyright (C) 1998,1999  Genome Research Limited
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * $Header: /nfs/disk222/yeastpub/Repository/powmap/embl/StreamQualifier.java,v 1.12 1999/04/27 14:53:37 kmr Exp $
 */

package embl;

import StringVector;

import java.io.BufferedReader;
import java.io.IOException;

/**
 *  This class contains routines for reading and writing Qualifiers.
 *
 *  @author Kim Rutherford
 *  @version $Id: StreamQualifier.java,v 1.12 1999/04/27 14:53:37 kmr Exp $
 **/

public // XXX

class StreamQualifier {
  /**
   *  Create a new Qualifier object by unquoting the value part of the
   *  Qualifier and the calling the Qualifier constructor.  This object
   *  consists of a name and a value.  In the raw embl file we have
   *  /name=value.
   *  @param name The name of this qualifier (ie. the text immediately after
   *    the / in the qualifier)
   *  @param value The value of this qualifier (ie the text immediately after
   *    the = in the qualifier).  This argument may be null if the qualifier
   *    has no value.  Unlike the Qualifier constructor the value String
   *    should include the quote characters (),"" or [] if the original
   *    qualifier contains them.  For example if the original qualifier was
   *    /citation=[3] then the value String should be: [3].
   *  @exception QualifierParseException Thrown if the value String is
   *    incorrectly quoted or formatted (see the Qualifier constructor).
   *  @exception InvalidQualifierException Thrown if the name String is not a
   *    valid embl qualifier name, if the value argument is null when there
   *    should be a value or if it has a value when it should be null
   **/
  public static Qualifier makeStreamQualifier (final String name,
                                               final String value)
      throws QualifierParseException, InvalidQualifierException {
    return new Qualifier (name, unquote (name, value));
  }

  /**
   *  Return a String version of the given Qualifier.
   **/
  public static String toString (final Qualifier qualifier) {
    final StringVector values = qualifier.getValues ();

    if (values == null) {
      return '/' + qualifier.getName ();
    } else {
      // the number we pick for the initial StringBuffer size is not critical,
      // but should cover most possibilities
      final StringBuffer buffer = new StringBuffer (50);

      for (int i = 0 ; i < values.size () ; ++i) {
        buffer.append ('/').append (qualifier.getName ());
        if (values.elementAt (i) != null) {
          buffer.append ('=').append (quotedValue (qualifier.getName (),
                                                   values.elementAt (i)));
        }
      }

      return buffer.toString ();
    }
  }

  /**
   *  Return a StringVector containing one String (of the form /name=value)
   *  for each of the values of the given qualifier.
   **/
  public static StringVector toStringVector (final Qualifier qualifier) {
    final StringVector values = qualifier.getValues ();

    final StringVector return_vector = new StringVector ();

    if (values == null) {
      return_vector.add ('/' + qualifier.getName ());
    } else {
      for (int i = 0 ; i < values.size () ; ++i) {
        // the number we pick for the initial StringBuffer size is not
        // critical
        final StringBuffer buffer = new StringBuffer (50);

        buffer.append ('/').append (qualifier.getName ());
        if (values.elementAt (i) != null) {
          buffer.append ('=').append (quotedValue (qualifier.getName (),
                                                   values.elementAt (i)));
        }
        return_vector.add (buffer.toString ());
      }
    }

    return return_vector;
  }

  /**
   *  This is used by readFromStream () as temporary storage.  It is a class
   *  member rather than a local variable so that we don't need to allocate a
   *  object for each call.  The number we pick for the initial StringBuffer
   *  size is not critical, but should cover most possibilities to prevent
   *  reallocation.
   **/
  final private static StringBuffer read_name_string_buffer =
    new StringBuffer (20);

  /**
   *  Read a qualifier name from a stream.
   *  @param buffered_reader the stream to read from
   *  @return the qualifier name if successful, otherwise null
   */
  static String readName (final BufferedReader buffered_reader)
      throws QualifierParseException, IOException {

    //System.out.println ("in readName ()");

    int current_char;

    while ((current_char = buffered_reader.read ()) != -1 &&
           0 != current_char    // Kaffe 1.00 returns 0 at end of string
           ) {
      if (' ' == current_char ||
          '\n' == current_char ||
          '\t' == current_char) {
        // read a whitespace character so go back to the top of the loop
        continue;
      } else {
        if ('/' == current_char) {
          // we have found the start of the qualifier name
          break;
        } else {
          // if the character isn't a / or space then something is wrong
          throw new QualifierParseException ("failed to read a qualifier " +
                                             "name from this string: " +
                                             (char)current_char +
                                             buffered_reader.readLine ());
        }
      }
    }

    if (-1 == current_char ||
        0 == current_char       // Kaffe 1.00 returns 0 at end of string
        ) {
      // end of file
      return null;
    }

    buffered_reader.mark (1);

    read_name_string_buffer.setLength (0);

    // the number we pick for the initial StringBuffer size is not critical,
    // but should cover most possibilities so that the buffer does not have to
    // increase its size
//    read_name_string_buffer.ensureCapacity (20);


    while ((current_char = buffered_reader.read ()) != -1) {
      if (Character.isLetter ((char) current_char) || '_' == current_char) {

        read_name_string_buffer.append ((char) current_char);

        // save the new position and go around the loop again
        buffered_reader.mark (1);
        continue;
      } else {
        // we have read one character too many
        buffered_reader.reset ();
        break;
      }
    }

    final String return_string = read_name_string_buffer.toString ();

    //System.out.println ("end of readName (): " + read_name_string_buffer.toString ());

    if (return_string.length () == 0) {
      throw new QualifierParseException ("zero length qualifier name read " +
                                         "from this string: " +
                                         buffered_reader.readLine ());
    } else {
      return return_string;
    }
  }

  /**
   *  This is used by readFromStream () as temporary storage.  It is a class
   *  member rather than a local variable so that we don't need to allocate a
   *  object for each call.  The number we pick for the initial array
   *  size is not critical, but should cover most possibilities to prevent
   *  reallocation.
   **/
  private static char [] read_value_buffer = new char [5000];

  /**
   *  The index into read_value_buffer - used by readValue () to keep track of
   *  where to put the next character.
   **/
  private static int buffer_index = 0;

  /**
   *  Read a qualifier value from a stream.
   *  @param buffered_reader the stream to read from
   *  @return the qualifier name if successful, otherwise null
   */
  static synchronized String readValue (final BufferedReader buffered_reader)
      throws QualifierParseException, IOException {

    buffer_index = 0;

    //System.out.println ("in readValue ()");

    buffered_reader.mark (1);

    int current_char = buffered_reader.read ();

    if (-1 == current_char) {
      throw new QualifierParseException ("hit the end of line while trying " +
                                         "to read a qualifier value");
    }

    // this is the character the marks the end of the value string.  the
    // default value of 0 means a '/' should end the string.
    char final_char = 0;

    if ('"' == current_char) {
      final_char = '"';
    }
    if ('[' == current_char) {
      final_char = ']';
    }
    if ('(' == current_char) {
      final_char = ')';
    }

    if (0 == final_char) {
      // the character we read isn't one of the delimiter characters so put it
      // back
      buffered_reader.reset ();
    } else {
      // append the char now so that loop doesn't stop immediately in the '"'
      // case
      appendToValueBuffer ((char) current_char);
    }

    // System.out.println ("in readValue () final_char = " + final_char);

    buffered_reader.mark (1);

    while ((current_char = buffered_reader.read ()) != -1) {

      // change newlines to spaces
      if (current_char == '\n') {
        current_char = ' ';
      }

      //System.out.println ("in readValue () read: " + (char) current_char);

      if (current_char == final_char) {
        // end of string
        appendToValueBuffer ((char) current_char);
        break;
      } else {
        if (0 == final_char && '/' == current_char) {
          // in this case '/' marks the end of the value. we need to push back
          // the '/' so that reading the next qualifier will work
          buffered_reader.reset ();
          break;
        } else {
          appendToValueBuffer ((char) current_char);

          // save the new position and go around the loop again
          buffered_reader.mark (1);
          continue;
        }
      }
    }

//    System.out.println ("end of readValue (): " + read_value_buffer.toString ());

    // move buffer_index back past any whitespace
    while (buffer_index > 0 &&
           Character.isWhitespace (read_value_buffer[buffer_index-1])) {
      --buffer_index;
    }

    return new String (read_value_buffer, 0, buffer_index);
  }

  /**
   *  Append the given char to read_value_buffer (at the position
   *  buffer_index), reallocating the buffer if necessary
   **/
  private static void appendToValueBuffer (final char new_char) {
    if (buffer_index >= read_value_buffer.length) {
      // reallocate as the buffer is full

      final char [] temp_buffer = new char [read_value_buffer.length*2];

      System.arraycopy (read_value_buffer, 0,
                        temp_buffer, 0,
                        read_value_buffer.length);
      read_value_buffer = temp_buffer;
    }

    read_value_buffer [buffer_index++] = (char) new_char;
  }

  /**
   *  Return the value part of a Qualifier correctly quoted for insertion into
   *  a embl entry.
   *  @param name The name part of qualifier.  The quote characters to check
   *    for depend on this name.
   *  @param value Quote this value.
   **/
  private static String quotedValue (final String name, final String value) {
    final QualifierInfo qualifier_info =
      EntryInformation.getQualifierInfo (name);

    if (qualifier_info == null) {
      throw new Error ("internal error - cannot find any type information " +
                       "for this qualifier name: " + name);
    }

    if (qualifier_info.getType () == QualifierInfo.QUOTED_TEXT ||
        qualifier_info.getType () == QualifierInfo.OPTIONAL) {
      return '"' + value + '"';
    } else {
      return value;
    }
  }

  /**
   *  Return the value part of a qualifier with any quote characters removed.
   *  @param name The name part of qualifier.  The quote characters to check
   *    for depend on this name.
   *  @param value This is the value String to strip the quote characters
   *    from.  This may be null if this qualifier has no value part (for
   *    example /partial).
   *  @exception QualifierParseException Thrown if the value String is
   *    incorrectly quoted for a qualifier with the given name.
   **/
  private static String unquote (final String name, final String value)
      throws QualifierParseException {
    final QualifierInfo qualifier_info =
      EntryInformation.getQualifierInfo (name);

    if (qualifier_info == null) {
      throw new QualifierParseException ("cannot find any type information " +
                                         "for this qualifier name: " + name);
    }

    if (qualifier_info.getType () == QualifierInfo.OPTIONAL &&
        value == null) {
      return null;
    }

    if (qualifier_info.getType () == QualifierInfo.NO_VALUE) {
      if (value == null) {
        return null;
      } else {
        throw new QualifierParseException ("this qualifier should not have " +
                                           "a value: " + name);
      }
    }

    if (value == null) {
      throw new QualifierParseException ("this qualifier should have " +
                                         "a value: " + name);
    }

    if (qualifier_info.getType () == QualifierInfo.QUOTED_TEXT ||
        qualifier_info.getType () == QualifierInfo.OPTIONAL) {
      // remove the quotes

      if (value.length () >= 2 &&
          value.charAt (0) == '"' &&
          value.charAt (value.length () - 1) == '"') {
        return value.substring (1, value.length () - 1);
      } else {
        throw new QualifierParseException ("this qualifier needs " +
                                           "double quotes around it's " +
                                           "value: " + name);
      }
    } else {
      if (value.length () > 0 && value.charAt (0) == '"') {
        throw new QualifierParseException ("the value of this qualifier " +
                                           "should not be quoted: " + name);
      }
    }

    // in the other cases leave the value as is
    return value;
  }
}
