/* StreamSequence.java (formally ReaderSequence.java)
 *
 * created: Wed Dec 30 1998
 *
 * This file is part of Artemis
 * 
 * Copyright (C) 1998,1999  Genome Research Limited
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * $Header: /nfs/disk222/yeastpub/Repository/powmap/embl/StreamSequence.java,v 1.12 1999/05/05 15:43:43 kmr Exp $
 */

package embl;

import java.io.IOException;
import java.io.Writer;

/**
 *  This is a subclass of Sequence that can read and write itself to a stream.
 *
 *  @author Kim Rutherford
 *  @version $Id: StreamSequence.java,v 1.12 1999/05/05 15:43:43 kmr Exp $
 **/

class StreamSequence
    extends LineGroup implements Sequence {
  /**
   *  Create a new StreamSequence object from a stream.  The next line to read
   *  from the stream should be the header line of the Sequence.
   *  @param in_stream The stream to read from.  When the constructor returns
   *    the stream will at the next line after the sequence.
   **/
  public StreamSequence (LinePushBackReader in_stream)
      throws IOException {

//    System.out.println ("in Sequence constructor");

    final String seq_header_line = in_stream.readLine ();


    if (seq_header_line.startsWith ("SQ   ")) {
      sequence_type = EMBL_FORMAT;
      in_stream.pushBack (seq_header_line);
      readEMBLSequence (in_stream);
    } else {
      if (seq_header_line.startsWith (">")) {
        sequence_type = FASTA_FORMAT;
        header_line = seq_header_line;
        readRawSequence (in_stream);
      } else {
        sequence_type = RAW_FORMAT;
        in_stream.pushBack (seq_header_line);
        readRawSequence (in_stream);
      }
    } 

    setCounts ();
  }

  /**
   *  Return the sequence type (one of EMBL_FORMAT, RAW_FORMAT, FASTA_FORMAT)
   **/
  public int getFormatType () {
    return sequence_type;
  }
  
  /**
   *  This is a possible value of the sequence_type variable which means the
   *  sequence was read in EMBL format.
   **/
  final public static int EMBL_FORMAT = 1;

  /**
   *  This is a possible value of the sequence_type variable which means the
   *  sequence was read in raw format.
   **/
  final public static int RAW_FORMAT = 2;

  /**
   *  This is a possible value of the sequence_type variable which means the
   *  sequence was read in fasta or similar format.  In this case the header
   *  line will be saved in the header_line member variable.
   **/
  final public static int FASTA_FORMAT = 3;

  /**
   *  This is used by the constructor as temporary storage.  It is a class
   *  member rather than a local variable so that we don't need to allocate a
   *  object for each call.
   **/
  final private StringBuffer this_line_sequence_buffer = new StringBuffer (90);

  /**
   *  This is used by writeToStream (), writeRaw (), etc as temporary storage.
   *  It is a class member rather than a local variable so that we don't need
   *  to allocate a object for each call.
   **/
  final private StringBuffer line_buffer = new StringBuffer (90);

  /**
   *  Write this Sequence to the given stream.  The output will be in whatever
   *  format the sequence was read in.
   *  @param writer The stream to write to.
   *  @param diana_extensions If true then the sequence will be written out in
   *    the format it was read, otherwise it will be written in EMBL format.
   **/
  public void writeToStream (final Writer writer,
                             final boolean diana_extensions)
      throws IOException {

    if (diana_extensions) {
      switch (sequence_type) {
      case RAW_FORMAT:
        writeRaw (writer);
        break;
      case EMBL_FORMAT:
        writeEMBL (writer);
        break;
      case FASTA_FORMAT:
        writeFasta (writer);
        break;
      default:
        throw new Error ("internal error - unknown sequence type");
      }
    } else {
      // write as embl
      writeEMBL (writer);
    }
  }

  /**
   *  Write this sequence to the given stream in raw format.
   *  @param writer The stream to write to.
   **/
  private void writeRaw (final Writer writer)
      throws IOException {
    final String sequence = toString ();

    final int SEQUENCE_LINE_BASE_COUNT = 60;

    for (int i = 0 ; i < sequence.length () ; i += SEQUENCE_LINE_BASE_COUNT) {

      // get the bases in chunks of at most 60

      final int this_line_length;

      if (sequence.length () - i < SEQUENCE_LINE_BASE_COUNT) {
        this_line_length = sequence.length () - i;
      } else {
        this_line_length = SEQUENCE_LINE_BASE_COUNT;
      }

      line_buffer.setLength (0);
      line_buffer.ensureCapacity (60);

      line_buffer.append (sequence.substring (i, i + this_line_length));

      line_buffer.append ("\n");

      writer.write (line_buffer.toString ());
    }
  }
  

  /**
   *  Write this Sequence to the given stream.  The output will be in FASTA
   *  format.
   *  @param writer The stream to write to.
   **/
  private void writeFasta (final Writer writer)
      throws IOException {
    writer.write (header_line + "\n");

    writeRaw (writer);
  }
  
  /**
   *  Write this Sequence to the given stream.  The output will be in EMBL
   *  format.
   *  @param writer The stream to write to.
   **/
  private void writeEMBL (final Writer writer)
      throws IOException {
    final String sequence = toString ();

    // first count A,C,G,T and other bases

    final int SEQUENCE_LINE_BASE_COUNT = 60;

    writer.write ("SQ   Sequence " +
                  sequence.length () + " BP; " +
                  getACount () + " A; " +
                  getCCount () + " C; " +
                  getGCount () + " G; " +
                  getTCount () + " T; " +
                  getOtherCount () + " other;\n");

    for (int i = 0 ; i < sequence.length () ; i += SEQUENCE_LINE_BASE_COUNT) {

      // get the bases in chunks of at most 60

      final int this_line_length;

      if (sequence.length () - i < SEQUENCE_LINE_BASE_COUNT) {
        this_line_length = sequence.length () - i;
      } else {
        this_line_length = SEQUENCE_LINE_BASE_COUNT;
      }

      line_buffer.setLength (0);
      line_buffer.ensureCapacity (90);

      line_buffer.append ("    ");

      final int BLOCK_LENGTH = 10;

      final int BLOCK_COUNT = 6;

      for (int j = 0 ; j < this_line_length ; j += BLOCK_LENGTH) {
        final int this_block_length;

        line_buffer.append (' ');

        if (this_line_length - j < BLOCK_LENGTH) {
          this_block_length = this_line_length - j;
        } else {
          this_block_length = BLOCK_LENGTH;
        }

        line_buffer.append (sequence.substring (i + j,
                                                i + j + this_block_length));
      }

      // the base counter to write at the end of each line
      final int base_count = i + this_line_length;

      final String string_base_count = String.valueOf (base_count);

      // now pad the line with spaces
      final int count_width = string_base_count.length ();

      final int LINE_WIDTH = 80;

      final int buffer_line_length = line_buffer.length ();

      for (int char_index = 0 ;
           char_index < LINE_WIDTH - count_width - buffer_line_length ;
           ++char_index) {
        line_buffer.append (' ');
      }

      line_buffer.append (string_base_count).append ("\n");

      writer.write (line_buffer.toString ());
    }
  }

  /**
   *  Return the type of this LineGroup object (SEQUENCE, FEATURE_TABLE, etc.)
   *  Returns SEQUENCE for objects of this class.
   **/
  public int getType () {
    return SEQUENCE;
  }
  
  /**
   *  Return the contents of this Sequence as a String.  This currently does
   *  not return a copy so any changes will change the Sequence object.
   **/
  public String toString () {
    return sequence;
  }

  /**
   *  Return a the given range of bases as a String.  Returns an empty
   *  sequence if the end position is less than the start position.
   *  @param start The start base of the range.
   *  @param end The end base of the range.
   **/
  public String getSubSequence (int start, int end) {
    if (end < start) {
      // empty range
      return "";
    } else {
      if (start == 1 && end == length ()) {
        return sequence;
      }

      // sanity checks
      if (end > length ()) {
        end = length ();
      }

      if (start < 1) {
        start = 1;
      }

      return sequence.substring (start - 1, end);
    }
  }

  /**
   *  Set this sequence to hold the bases in the given String.
   **/
  public void setFromString (final String new_sequence) {
    sequence = new_sequence;
    setCounts ();
  }

  /**
   *  Returns the length of the sequence in bases.
   **/
  public int length () {
    return sequence.length ();
  }

  /**
   *  Return the count of c bases in the whole sequence.
   **/
  public int getCCount () {
    return c_count;
  }

  /**
   *  Return the count of g bases in the whole sequence.
   **/
  public int getGCount () {
    return g_count;
  }

  /**
   *  Return the count of a bases in the whole sequence.
   **/
  public int getACount () {
    return a_count;
  }

  /**
   *  Return the count of t bases in the whole sequence.
   **/
  public int getTCount () {
    return t_count;
  }

  /**
   *  Return the count of non-g,c,t,a bases in the whole sequence.
   **/
  public int getOtherCount () {
    return
      length () - (getCCount () + getACount () + getTCount () + getGCount ());
  }

  /**
   *  Take a EMBL sequence header line and extract and return the base count.
   *  If the line cannot be parsed as a header line then -1 is returned.
   **/
  private int getHeaderBaseCount (final String line) {
    System.out.println ("sequence header: " + line);

    if (line.startsWith ("SQ   Sequence ")) {
      final String temp_line = line.substring (14);

      final int space_index = temp_line.indexOf (' ');

      if (space_index == -1) {
        return -1;
      }

      final String count_string = temp_line.substring (0, space_index);

      try {
        return Integer.parseInt (count_string);
      } catch (NumberFormatException e) {
        return -1;
      }
    } else {
      return -1;
    }
  }

  /**
   *  Set the a_count, c_count, t_count and g_count variables.
   **/
  private void setCounts () {
    a_count = c_count = t_count = g_count = 0;

    for (int i = 0 ; i < sequence.length () ; ++i) {
      switch (sequence.charAt (i)) {
      case 'a':
        ++a_count;
        break;
      case 'c':
        ++c_count;
        break;
      case 'g':
        ++g_count;
        break;
      case 't':
        ++t_count;
        break;
      default:
        break;
      }
    }
  }


  /**
   *  Read EMBL sequence from the given stream.
   **/
  private void readEMBLSequence (final LinePushBackReader in_stream)
      throws IOException {
    final String seq_header_line = in_stream.readLine ();

    final int header_base_count = getHeaderBaseCount (seq_header_line);

    final int buffer_capacity;

    if (header_base_count > 50000) {
      // add 100 for good luck
      buffer_capacity = header_base_count + 100;
    } else {
      // try to pick an initial that covers a large number of cases
      buffer_capacity = 50000;
    }

    // we buffer up the sequence bases then assign them to sequence once all
    // bases are read
    StringBuffer sequence_buffer = new StringBuffer (buffer_capacity);

    while (true) {
      // temporary storage for reading lines
      String line = in_stream.readLine ();

      if (line == null) {
        // we have reached the end of file - break out of loop
        break;
      }

      if (getLineType (line) != SEQUENCE) {
        // end of Sequence -
        in_stream.pushBack (line);
        break;
      }

      if (line.length () < 80) {
        throw new ReadFormatException ("line too short while reading " +
                                       "sequence data",
                                       in_stream.getLineNumber ());
      }


      this_line_sequence_buffer.setLength (0);
      this_line_sequence_buffer.ensureCapacity (60);

      // if the input file is a well formatted embl entry the bases should be
      // in 6 columns of 10 bases each (with a single space separating them).
      // the first column starts at index 5 of the input line.

      final char [] this_line_char_array = line.toCharArray ();

      for (int i = 0 ; i < 6 ; ++i) {
        final int first_base = 5 + i * 11;

        this_line_sequence_buffer.append (this_line_char_array,
                                          first_base,
                                          10);
      }

      // add the sequence from this line to the sequence we already have.

      String this_line_sequence =
        this_line_sequence_buffer.toString ();

      if (this_line_char_array[59] == ' ') {
        // call trim () because this line (the last line) ends in spaces.
        sequence_buffer.append (this_line_sequence.trim ().toLowerCase ());
      } else {
        sequence_buffer.append (this_line_sequence.toLowerCase ());
      }
    }

    sequence = sequence_buffer.toString ();
  }
  
  /**
   *  This method will read raw sequence from the stream in this object
   *  removing whitespace as it goes.  No checks are made on the format of the
   *  sequence.
   **/
  private void readRawSequence (final LinePushBackReader in_stream)
      throws IOException {

    final int buffer_capacity = 50000;

    // we buffer up the sequence bases then assign them to sequence once all
    // bases are read
    StringBuffer sequence_buffer = new StringBuffer (buffer_capacity);

    while (true) {
      // temporary storage for reading lines
      String line = in_stream.readLine ();

      if (line == null) {
        // we have reached the end of file - break out of loop
        break;
      }

      line = line.toLowerCase ();
      
      if (line.equals ("*")) {
        // end of the sequence
        break;
      }

      if (line.startsWith (">")) {
//        sequence_buffer.append ('@');

        // ignore header lines
        continue;
      }
      
      for (int i = 0 ; i < line.length () ; ++i) {
        final char this_char = line.charAt (i);

        if (Character.isLetter (this_char) ||
            this_char == '-' ||
            this_char == '*') {
          sequence_buffer.append (this_char);
        } else {
          if (Character.isSpaceChar (this_char)) {
            // just ignore it
          } else {
            throw new ReadFormatException ("sequence file contains a " +
                                           "character that is not a letter: " +
                                           this_char,
                                           in_stream.getLineNumber ());
          }
        }
      }
    }
    
    sequence = sequence_buffer.toString ();
  }

  /**
   *  Contains the sequence data for this object.  It will contain the bases
   *  of the sequence with no spaces after the Feature constructor finishes.
   **/
  private String sequence;

  /**
   *  Count of the a bases in the sequence.
   **/
  private int a_count = 0;

  /**
   *  Count of the c bases in the sequence.
   **/
  private int c_count = 0;

  /**
   *  Count of the g bases in the sequence.
   **/
  private int g_count = 0;

  /**
   *  Count of the t bases in the sequence.
   **/
  private int t_count = 0;

  /**
   *  The format type of the sequence (as read).
   **/
  private int sequence_type = RAW_FORMAT;

  /**
   *  The header line read from the input stream or null if the input was
   *  raw.  This is used when reading FASTA format.
   **/
  private String header_line = null;
}
